
/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of GpxParser.
 *  
 *  GpxParser is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  GpxParser is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with GpxParser.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 

import java.util.Vector;
import java.io.FileWriter;
import java.io.BufferedWriter;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;

public class GpxParser extends DefaultHandler {

  /*** member variables ***/
  private Vector<TrackPoint> m_pTrackPoints;
  private Vector<Target> m_pTargets;
  private TrackPoint m_pTrackPointToCreate;
  private boolean m_bElevationStarted, m_bTimeStarted, m_bHoursOffsetApplied;
  private Configuration m_pConfiguration;
  
  /////////////////////////////////
  // CONSTRUCTOR                 //
  /////////////////////////////////
  public GpxParser(Configuration pConfiguration) throws Exception {
    if (pConfiguration == null) throw new NullPointerException("No configuration has been specified!");
    m_pConfiguration = pConfiguration;
    
    m_bHoursOffsetApplied = false;
    
    
    m_pTrackPoints = new Vector<TrackPoint>();
    m_pTargets = new Vector<Target>();
    m_pTrackPointToCreate = null;
    m_bElevationStarted = false;
    m_bTimeStarted = false;
    
    //parse the GPX
    SAXParserFactory pFactory = SAXParserFactory.newInstance();
    pFactory.setNamespaceAware(true);
    SAXParser pSaxParser = pFactory.newSAXParser();
    XMLReader pReader = pSaxParser.getXMLReader();
    pReader.setContentHandler(this);
    pReader.parse(m_pConfiguration.getProperty("InputFile"));
    
  }
  
  
  ///////////////////////////////////////////
  // DEFAULT HANDLER (SAXPARSER) INTERFACE //
  ///////////////////////////////////////////
  
  public void startDocument() throws SAXException {
    //tasks when starting the xml parsing
  }
  
  public void endDocument() throws SAXException {
    int iTrackPointNum = m_pTrackPoints.size();
    TrackPoint pPreviousTrackPoint = null;
    for (int i = 0; i < iTrackPointNum; i++) {
      TrackPoint pCurrentTrackPoint = m_pTrackPoints.get(i);
      pCurrentTrackPoint.calculateValues(pPreviousTrackPoint);
      pPreviousTrackPoint = pCurrentTrackPoint;
    }
  }
  
  public void startElement(
    String pNamespaceURI,
    String pLocalName,
    String pQualifiedName, 
    Attributes pAttributes
  ) {
    if (pLocalName.equals("trkpt")) {
      m_pTrackPointToCreate = new TrackPoint(pAttributes.getValue("lat"), pAttributes.getValue("lon"));
    }
    else if (pLocalName.equals("ele")) {
      m_bElevationStarted = true;
    }
    else if (pLocalName.equals("time")) {
      m_bTimeStarted = true;
    }
    else if (pLocalName.equals("target") || pLocalName.equals("targetleave")) {
      String pName = pAttributes.getValue("name");
      String pTime = pAttributes.getValue("time");
      String pElevation = pAttributes.getValue("ele");
      if (pLocalName.equals("target")) {
        m_pTargets.add(new Target(pName, pTime, pElevation));
      }
      else {
        m_pTargets.add(new Target(pName, pTime, pElevation, true));
      }
    }
  
  }
  
  public void characters(char[] pCharacters, int iStart, int iLength) throws SAXException {
    if (m_bElevationStarted) {
      m_pTrackPointToCreate.setElevation(new String(pCharacters, iStart, iLength));
      m_bElevationStarted = false;
    }
    else if (m_bTimeStarted) {
      if (m_pTrackPointToCreate != null) {
        m_pTrackPointToCreate.setTimestamp(new String(pCharacters, iStart, iLength));
      }
      m_bTimeStarted = false;
    }
  }
  
  public void endElement(
    String pNamespaceURI,
    String pLocalName,
    String pQualifiedName
  ) {
    if (pLocalName.equals("trkpt")) {
      m_pTrackPoints.add(m_pTrackPointToCreate);
      //System.out.println(m_pTrackPointToCreate.toString());
      m_pTrackPointToCreate = null;
    }
  }
  
  //////////////////////////////////
  // CSV EXPORT                   //
  //////////////////////////////////
  public void writeHeightAnalysis() throws Exception {
    applyHoursOffset();
    HeightAnalysis pAnalysis = new HeightAnalysis(m_pTrackPoints, m_pTargets);
    Vector<String> pCsvData = pAnalysis.getCsvData();
  
    String pCsvFileName = m_pConfiguration.getProperty("HeightAnalysisOutput");
    BufferedWriter pWriter = new BufferedWriter(new FileWriter(pCsvFileName));
    int iLineNum = pCsvData.size();
    for (int i = 0; i < iLineNum; i++) {
      pWriter.write(pCsvData.get(i));
      pWriter.newLine();
    }
    
    pWriter.close();
  }
  
  public void writeTimeStatistic() throws Exception {
    applyHoursOffset();
    
    TimeStatistic pTimeStatistic = new TimeStatistic(m_pTargets, m_pTrackPoints);
    Vector<String> pHtml = pTimeStatistic.getHtml();
    int iHtmlLines = pHtml.size();
    
    
    String pHtmlFileName = m_pConfiguration.getProperty("TimeStatisticOutput");
    BufferedWriter pWriter = new BufferedWriter(new FileWriter(pHtmlFileName));
    
    for (int i = 0; i < iHtmlLines; i++) {
      pWriter.write(pHtml.get(i));
      pWriter.newLine();
    }
    
    pWriter.close();
    
    
    
  }
  

  
  //////////////////////////////////
  // PROGRAMM START               //
  //////////////////////////////////

  public static void main(String[] args) throws Exception {
    
    Configuration pConfiguration = new Configuration("GpxParser.properties");
    GpxParser pParser = new GpxParser(pConfiguration);
    
    pParser.writeHeightAnalysis();
    pParser.writeTimeStatistic();
  }
  
  
  ////////////////////////////////////
  // PRIVATE HELPER METHODS         //
  ////////////////////////////////////
  private void applyHoursOffset() {
    //1. some checks
    if (m_bHoursOffsetApplied) return;
    
    if (m_pTargets.size() == 0) return;
    if (m_pTrackPoints.size() == 0) return;
    
    Target pFirstTarget = m_pTargets.get(0);
    TrackPoint pFirstTrackPoint = m_pTrackPoints.get(0);
    
    
    //2. calculate the current time difference in milliseconds
    long lTargetTime = pFirstTarget.getTimeInMilliseconds();
    long lTrackPointTime = pFirstTrackPoint.getTimeInMilliseconds();
    
    long lDiff = lTargetTime - lTrackPointTime;
    
    
    //3. calculate the offset
    int iOffset = 0;
    
    int iOffsetStep = 1;
    if (lDiff < 0) iOffsetStep = -1;
    while (true) {
      long lCurrentOffsetDiff = Math.abs(lDiff - iOffset*1000*60*60);
      long lNextOffsetDiff = Math.abs(lDiff - (iOffset+iOffsetStep)*1000*60*60);
      
      if (lNextOffsetDiff > lCurrentOffsetDiff) break;

      iOffset++;
    }
    
    //4. apply the offset
    int iTrackPointNum = m_pTrackPoints.size();
    for (int i = 0; i < iTrackPointNum; i++) {
      m_pTrackPoints.get(i).applyHoursOffset(iOffset);
    }
    
    

    //5. finish
    System.out.println("[INFO] Hours offset " + iOffset + " applied!");
    m_bHoursOffsetApplied = true;
  }
  

}