
/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of GpxParser.
 *  
 *  GpxParser is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  GpxParser is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with GpxParser.  If not, see <http://www.gnu.org/licenses/>.
 */
 

import java.util.Vector;
import java.io.FileWriter;
import java.io.BufferedWriter;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;

public class Kml2Gpx extends DefaultHandler {

  /*** member variables ***/
  private Vector<TrackPoint> m_pTrackPoints;
  private TrackPoint m_pTrackPointToCreate;
  private Configuration m_pConfiguration;
  private boolean m_bWhenStarted, m_bCoordStarted;
  private String m_pAuthor, m_pName;
  
  /////////////////////////////////
  // CONSTRUCTOR                 //
  /////////////////////////////////
  public Kml2Gpx(Configuration pConfiguration) throws Exception {
    if (pConfiguration == null) throw new NullPointerException("No configuration has been specified!");
    m_pConfiguration = pConfiguration;
    
    
    
    m_pTrackPoints = new Vector<TrackPoint>();
    m_pTrackPointToCreate = null;
    m_bWhenStarted = false;
    m_bCoordStarted = false;
    m_pAuthor = "GpxParser - Kml2Gpx";
    m_pName = "";
    
    //parse the GPX
    SAXParserFactory pFactory = SAXParserFactory.newInstance();
    pFactory.setNamespaceAware(true);
    SAXParser pSaxParser = pFactory.newSAXParser();
    XMLReader pReader = pSaxParser.getXMLReader();
    pReader.setContentHandler(this);
    pReader.parse(m_pConfiguration.getProperty("InputFile"));
    
    //write the content to a new xml
    writeGpx();
    
  }
  
  
  ///////////////////////////////////////////
  // DEFAULT HANDLER (SAXPARSER) INTERFACE //
  ///////////////////////////////////////////
  
  public void startDocument() throws SAXException {
    //tasks when starting the xml parsing
  }
  
  public void endDocument() throws SAXException {
  }
  
  public void startElement(
    String pNamespaceURI,
    String pLocalName,
    String pQualifiedName, 
    Attributes pAttributes
  ) {
    if (pLocalName.equals("when")) {
      m_bWhenStarted = true;
    }
    else if (pLocalName.equals("coord")) {
      m_bCoordStarted = true;
    }
  }
  
  public void characters(char[] pCharacters, int iStart, int iLength) throws SAXException {
    if (m_bWhenStarted) {
      // date: 2013-05-19T08:01:59.825Z
      String pValue = new String(pCharacters).substring(iStart, iStart+iLength);
      m_pTrackPointToCreate = new TrackPoint(pValue);
      m_bWhenStarted = false;
      //System.out.println(pValue);  
      //System.out.println("when characters" + ", " + iStart + ", " + iLength);  
    }
    else if (m_bCoordStarted) {
      if (m_pTrackPointToCreate != null) {
        //content: 15.909191 47.574274 1407.9627685546875
        String pContent = new String(pCharacters).substring(iStart, iStart+iLength);
        String[] pSplitted = pContent.split(" ");
        if (pSplitted.length < 3) throw new SAXException("The content of the gx:coord entity is in the wrong format! " + pContent);
        
        m_pTrackPointToCreate.setLongitude(pSplitted[0]);
        m_pTrackPointToCreate.setLatitude(pSplitted[1]);
        m_pTrackPointToCreate.setElevation(pSplitted[2]);
        
        m_pTrackPoints.add(m_pTrackPointToCreate);
        m_pTrackPointToCreate = null;
      }
      m_bCoordStarted = false;
    }
  }
  
  public void endElement(
    String pNamespaceURI,
    String pLocalName,
    String pQualifiedName
  ) {
    
  }
  
  //////////////////////////////////
  // CSV EXPORT                   //
  //////////////////////////////////
  
  
  public void writeGpx() throws Exception {
    
    Vector<String> pXml = getGpxContent();
    int iXmlLines = pXml.size();
    
    
    String pGpxFile = m_pConfiguration.getProperty("OutputFile");
    BufferedWriter pWriter = new BufferedWriter(new FileWriter(pGpxFile));
    
    for (int i = 0; i < iXmlLines; i++) {
      pWriter.write(pXml.get(i));
      pWriter.newLine();
    }
    
    pWriter.close();
    
    
    
  }
  

  
  //////////////////////////////////
  // PROGRAMM START               //
  //////////////////////////////////

  public static void main(String[] args) throws Exception {
    
    Configuration pConfiguration = new Configuration("Kml2Gpx.properties");
    Kml2Gpx pConverter = new Kml2Gpx(pConfiguration);
    
  }
  
  
  ////////////////////////////////////
  // PRIVATE HELPER METHODS         //
  ////////////////////////////////////
  private Vector<String> getGpxContent() {
    Vector<String> pXml = new Vector<String>();
    pXml.add("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    pXml.add("<gpx version=\"1.1\" creator=\"" + m_pAuthor + "\">");
    pXml.add("  <trk>");
    pXml.add("    <name>" + m_pName + "</name>");
    pXml.add("    <trkseg>");
    
    //to do track points
    int iTrackPointNum = m_pTrackPoints.size();
    System.out.println(iTrackPointNum + " found!");
    for (int i = 0; i < iTrackPointNum; i++) {
      pXml.add(m_pTrackPoints.get(i).getGpx());
    }
    /*
			<trkpt lat="47.676143" lon="15.723030"><ele>1070.737549</ele><time>2017-09-09T08:03:57Z</time></trkpt>
    */
    
    pXml.add("    </trkseg>");
    pXml.add("  </trk>");
    pXml.add("</gpx>");
    return pXml;
  }
  
  

}