
/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of GpxParser.
 *  
 *  GpxParser is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  GpxParser is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with GpxParser.  If not, see <http://www.gnu.org/licenses/>.
 */
 

import java.util.Vector;
import java.util.Locale;
import java.text.DecimalFormatSymbols;
import java.text.DecimalFormat;

public class TimeStatistic {

  /*** constants ***/
  private static final DecimalFormat KM_FORMAT = new DecimalFormat("#0.00", new DecimalFormatSymbols(Locale.GERMAN));

  /*** member variables ***/
  private Vector<Target> m_pTargets;
  private Vector<TrackPoint> m_pTrackPoints;
  private int m_iTotalMinutes; 
  private double m_dTotalKilometers;
  private boolean m_bDirectDistances;
  
  ///////////////////////////////
  // CONSTRUCTOR               //
  ///////////////////////////////
  public TimeStatistic(Vector<Target> pTargets, Vector<TrackPoint> pTrackPoints) {
    if (pTargets == null) throw new NullPointerException("No targets have been defined!");
    if (pTrackPoints == null) throw new NullPointerException("No track points have been defined!");
    
    m_pTargets = pTargets;
    m_pTrackPoints = pTrackPoints;
    
    m_iTotalMinutes = 0;
    m_dTotalKilometers = 0;
    m_bDirectDistances = false;
  }
  
  /////////////////////////////////
  // GET METHODS                 //
  /////////////////////////////////
  public Vector<String> getHtml() {
    Vector<String> pHtml = new Vector<String>();
    pHtml.add("<html>");
    
    m_bDirectDistances = false;
    pHtml.add("  <b>GPS Distanzen</b><br/><br/>");
    pHtml.addAll(getTableHtml());
    
    pHtml.add("  <br/><br/>");
    
    m_bDirectDistances = true;
    pHtml.add("  <b>Direkte Distanzen (H&ouml;henmeter inkludiert)</b><br/><br/>");
    pHtml.addAll(getTableHtml());
    
    pHtml.add("</html>");
    
    return pHtml;
  }
  
  ////////////////////////////////
  // PRIVATE HELPERS            //
  ////////////////////////////////
  private Vector<String> getTableHtml() {
    Vector<String> pHtml = new Vector<String>();
      
    //reset the member variables for a new calculation
    m_iTotalMinutes = 0;
    m_dTotalKilometers = 0;
  
    
    pHtml.add("  <table class='timing'>");
    
    pHtml.addAll(getHeaderHtml());
    
    int iTargetNum = m_pTargets.size();
    Target pPreviousTarget = null;
    for (int i = 0; i < iTargetNum; i++) {
      Target pCurrTarget = m_pTargets.get(i);
      pHtml.addAll(getRowHtml(pCurrTarget, pPreviousTarget));
      pPreviousTarget = pCurrTarget;
    }
    
    pHtml.add("  </table>");
    
    return pHtml;
    
  }
  
  
  private Vector<String> getHeaderHtml() {
    Vector<String> pHtml = new Vector<String>();
    pHtml.add(getRowStart());
    pHtml.add(getColumn("Ort", true, true));
    pHtml.add(getColumn("Zeit", false, true));
    pHtml.add(getColumn("Dauer<br/>Abschnitt", false, true));
    pHtml.add(getColumn("km<br/>Abschnitt", false, true));
    pHtml.add(getColumn("km/h<br/>Abschnitt", false, true));
    pHtml.add(getColumn("Dauer<br/>Gesamt", false, true));
    pHtml.add(getColumn("km<br/>Gesamt", false, true));
    pHtml.add(getColumn("km/h<br/>Gesamt", false, true));
    
    pHtml.add(getRowEnd());
    return pHtml;
  }
  
  private Vector<String> getRowHtml(Target pTarget, Target pPreviousTarget) {
    Vector<String> pHtml = new Vector<String>();
    if (pTarget == null) return pHtml;
    
    //calculations
    int iSectionMinutes = 0;
    double dSectionKilometers = 0;
    if ((pPreviousTarget != null) && (!pTarget.isTargetLeave())) {
      iSectionMinutes = pTarget.getAbsoluteMinuteDifference(pPreviousTarget);
      m_iTotalMinutes += iSectionMinutes;
      double dNewTotalKilometers = getAbsoluteKilometersOfTarget(pTarget);
      dSectionKilometers = dNewTotalKilometers - m_dTotalKilometers;
      m_dTotalKilometers = dNewTotalKilometers;
    }
  
    
    pHtml.add(getRowStart());
    pHtml.add(getColumn(pTarget.getName() + "<br/>" + pTarget.getElevationForHtml(), true, false));
    pHtml.add(getColumn(pTarget.getTimeForHtml(), false, false));
    if (m_iTotalMinutes == 0) {
      pHtml.add(getColumn("-", false, false));
      pHtml.add(getColumn("-", false, false));
      pHtml.add(getColumn("-", false, false));
      pHtml.add(getColumn("-", false, false));
      pHtml.add(getColumn("-", false, false));
      pHtml.add(getColumn("-", false, false));
    }
    else {
      pHtml.add(getColumn(getDurationText(iSectionMinutes), false, false));
      pHtml.add(getColumn(getKilometerText(dSectionKilometers), false, false));
      pHtml.add(getColumn(getSpeedText(dSectionKilometers, iSectionMinutes), false, false));
      pHtml.add(getColumn(getDurationText(m_iTotalMinutes), false, false));
      pHtml.add(getColumn(getKilometerText(m_dTotalKilometers), false, false));
      pHtml.add(getColumn(getSpeedText(m_dTotalKilometers, m_iTotalMinutes), false, false));
    }
    
    pHtml.add(getRowEnd());
    return pHtml;
  }
  
  private String getRowStart() {
    return "    <tr>";
  }
  
  private String getRowEnd() {
    return "    </tr>";
  }
  
  private String getColumn(String pValue, boolean bLocation, boolean bBold) {
    pValue = Utilities.prepareString(pValue);
    if (bBold) pValue = "<b>" + pValue + "</b>";
    
    String pClass = "timing";
    if (bLocation) pClass = "timingLocation";
    
    String pHtml = "      <td class='" + pClass + "'>" + pValue + "</td>";
    return pHtml;
  }
  
  private String getDurationText(int iTotalMinutes) {
    if (iTotalMinutes <= 0) return "-";
    
    int iHours = (int)(iTotalMinutes/60);
    int iMinutes = iTotalMinutes - 60*iHours;
    
    String pResult = "";
    if (iHours > 0) {
      pResult += iHours + " h ";
      if (iMinutes < 10) pResult += "0";
    }
    pResult += iMinutes + " m";
    return pResult;
  }
  
  private String getKilometerText(double dKilometers) {
    if (dKilometers <= 0) return "-";
    
    return KM_FORMAT.format(dKilometers);
  }
  
  private String getSpeedText(double dKilometers, int iMinutes) {
    if (iMinutes <= 0) return "-";
    
    return KM_FORMAT.format(dKilometers*60/iMinutes);
  }
  
  
  private double getAbsoluteKilometersOfTarget(Target pTarget) {
    //1. do a simple check
    if (pTarget == null) return 0;
    
    
    
    //2. find a track point with the exact time
    String pTargetTime = pTarget.getTimeForHtml();
    int iTrackPointNum = m_pTrackPoints.size();
    
    for (int i = 0; i < iTrackPointNum; i++) {
      TrackPoint pTrackPoint = m_pTrackPoints.get(i);
      String pTime = pTrackPoint.getTimeForHtml();
      if (pTime.equals(pTargetTime)) return getKilometersFromTrackPoint(pTrackPoint);
      
    }
    
    //3. when there is no track point with the exact time, find the nearest track point
    long lTargetTimestamp = pTarget.getTimeInMilliseconds();
    
    for (int i = 0; i < iTrackPointNum-1; i++) {
      TrackPoint pTrackPoint = m_pTrackPoints.get(i);
      TrackPoint pNextTrackPoint = m_pTrackPoints.get(i+1);
      
      long lTrackPointTimestamp = pTrackPoint.getTimeInMilliseconds();
      long lNextTrackPointTimestamp = pNextTrackPoint.getTimeInMilliseconds();
      
      
      if ((lTargetTimestamp >= lTrackPointTimestamp) && (lTargetTimestamp < lNextTrackPointTimestamp)) {
        
        //System.out.println(lTrackPointTimestamp + " - " + lTargetTimestamp + " - " + lNextTrackPointTimestamp);
        
        long lDiff1 = Math.abs(lTrackPointTimestamp - lTargetTimestamp);
        long lDiff2 = Math.abs(lNextTrackPointTimestamp - lTargetTimestamp);
        if (lDiff1 < lDiff2) {
          //System.out.println("Target: " + pTargetTime + ", Trackpoint: " + pTrackPoint.getTimeForHtml());
          return getKilometersFromTrackPoint(pTrackPoint);
        }
        else {
          //System.out.println("Target: " + pTargetTime + ", Trackpoint: " + pNextTrackPoint.getTimeForHtml());
          return getKilometersFromTrackPoint(pNextTrackPoint);
        }
      }
    }
    
    //4. if nothing is found return the last track point.
    return getKilometersFromTrackPoint(m_pTrackPoints.get(iTrackPointNum-1));
  }
  
  private double getKilometersFromTrackPoint(TrackPoint pTrackPoint) {
    if (m_bDirectDistances) {
      return pTrackPoint.getDirectDistance();
    }
    else {
      return pTrackPoint.getKilometers();
    }
  }
  
  
  /*
    
    <table class="timing">
        <tr>
          <td class="timingLocation"><b>Ort</b></td>
          <td class="timing"><b>Zeit</b></td>
          <td class="timing"><b>Dauer<br/>Abschnitt</b></td>
          <td class="timing"><b>km<br/>Abschnitt</b></td>
          <td class="timing"><b>km/h<br/>Abschnitt</b></td>
          <td class="timing"><b>Dauer<br/>Gesamt</b></td>
          <td class="timing"><b>km<br/>Gesamt</b></td>
          <td class="timing"><b>km/h<br/>Gesamt</b></td>
        </tr>
        <tr>
          <td class="timingLocation">Steyersberger Schwaig<br/>1364 m</td>
          <td class="timing">10:30</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
        </tr>
        <tr>
          <td class="timingLocation">Feistritzer Schwaig Ankunft<br/>1443 m</td>
          <td class="timing">11:41</td>
          <td class="timing">1 h 11 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
          <td class="timing">1 h 11 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
        </tr>
        <tr>
          <td class="timingLocation">Feistritzer Schwaig Aufbruch<br/>1443 m</td>
          <td class="timing">12:15</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">1 h 11 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
        </tr>
        <tr>
          <td class="timingLocation">Dreiländereck<br/>1561 m</td>
          <td class="timing">13:08</td>
          <td class="timing">53 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
          <td class="timing">2 h 03 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
        </tr>
        <tr>
          <td class="timingLocation">Kranichberger Schwaig Ankunft<br/>1535 m</td>
          <td class="timing">13:19</td>
          <td class="timing">11 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
          <td class="timing">2 h 14 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
        </tr>
        <tr>
          <td class="timingLocation">Kranichberger Schwaig Aufbruch<br/>1535 m</td>
          <td class="timing">13:48</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">-</td>
          <td class="timing">2 h 14 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
        </tr>
        <tr>
          <td class="timingLocation">Steyersberger Schwaig<br/>1364 m</td>
          <td class="timing">14:28</td>
          <td class="timing">40 m</td>
          <td class="timing">?</td>
          <td class="timing">?</td>
          <td class="timing">2 h 54 m</td>
          <td class="timing">10,05</td>
          <td class="timing">3,47</td>
        </tr>
        
        
      </table>
      */

}