
/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of GpxParser.
 *  
 *  GpxParser is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  GpxParser is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with GpxParser.  If not, see <http://www.gnu.org/licenses/>.
 */


import java.util.Date;
import java.text.SimpleDateFormat;

public class TrackPoint {

  /*** constants ***/
  private static final SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("HH:mm");
  private static final double EARTH_RADIUS = 6371; 

  /*** member variables ***/
  private double m_dLatitude, m_dLongitude, m_dElevation, m_dKilometers, m_dDirectDistance;
  private Date m_pTimestamp;
  private long m_lTimeInMilliseconds;
  
  
  ///////////////////////////////
  // CONSTRUCTOR               //
  ///////////////////////////////
  public TrackPoint(String pLatitude, String pLongitude) {
    m_dLatitude = Utilities.getDouble(pLatitude);
    m_dLongitude = Utilities.getDouble(pLongitude);
    m_dElevation = 0;
    m_pTimestamp = null;
    m_lTimeInMilliseconds = 0;
    m_dKilometers = 0;
    m_dDirectDistance = 0;
  }
  
  public TrackPoint(String pKmlTimestamp) {
    m_dLatitude = 0;
    m_dLongitude = 0;
    m_dElevation = 0;
    //System.out.println(pKmlTimestamp + " vs. " + Utilities.getDateFromKml(pKmlTimestamp));
    m_pTimestamp = Utilities.getDateFromKml(pKmlTimestamp);
    m_lTimeInMilliseconds = Utilities.getTimeInMilliseconds(m_pTimestamp);
    m_dKilometers = 0;
    m_dDirectDistance = 0;
  }
  
  
  ///////////////////////////////
  // SET METHODS               //
  ///////////////////////////////
  public void applyHoursOffset(int iHoursOffset) {
    m_pTimestamp = Utilities.addHours(m_pTimestamp, iHoursOffset);
    m_lTimeInMilliseconds = Utilities.getTimeInMilliseconds(m_pTimestamp);
  }
  
  
  public void setElevation(String pElevation) {
    double dElevation = Utilities.getDouble(pElevation);
    if (dElevation != Double.NaN) m_dElevation = dElevation;
    
  }
  
  public void setTimestamp(String pTimestamp) {
    m_pTimestamp = Utilities.getDateFromXml(pTimestamp);
    m_lTimeInMilliseconds = Utilities.getTimeInMilliseconds(m_pTimestamp);
  }
  
  public void setLatitude(String pLatitude) {
    m_dLatitude = Utilities.getDouble(pLatitude);
  }
  
  public void setLongitude(String pLongitude) {
    m_dLongitude = Utilities.getDouble(pLongitude);
  }
  
  public void calculateValues(TrackPoint pPreviousTrackPoint) {
    if (pPreviousTrackPoint == null) {
      m_dKilometers = 0;
      m_dDirectDistance = 0;
    }
    else {
      //source of the code: 
      //http://stackoverflow.com/questions/639695/how-to-convert-latitude-or-longitude-to-meters
      //haversine equation: https://en.wikipedia.org/wiki/Haversine_formula
      double dPrevLatitude = pPreviousTrackPoint.getLatitude();
      double dPrevLongitude = pPreviousTrackPoint.getLongitude();
      
      double dRadiansFactor = Math.PI / 180;
      double dLatitudeDiff = m_dLatitude*dRadiansFactor - dPrevLatitude*dRadiansFactor;
      double dLongitudeDiff = m_dLongitude*dRadiansFactor - dPrevLongitude*dRadiansFactor;
      
      
      double dA = Math.sin(dLatitudeDiff/2) * Math.sin(dLatitudeDiff/2) +
                  Math.cos(m_dLatitude * dRadiansFactor) * Math.cos(dPrevLatitude * dRadiansFactor) *
                  Math.sin(dLongitudeDiff/2) * Math.sin(dLongitudeDiff/2);
      double dC = 2 * Math.atan2(Math.sqrt(dA), Math.sqrt(1-dA));
      double dFlatKilometers = EARTH_RADIUS * dC;
      
      m_dKilometers = dFlatKilometers + pPreviousTrackPoint.getKilometers();
      
      double dHeightDiff = (m_dElevation - pPreviousTrackPoint.getElevation())/1000; //elevation in m -> /1000 -> in km
      m_dDirectDistance = Math.sqrt(dFlatKilometers*dFlatKilometers + dHeightDiff*dHeightDiff)  + pPreviousTrackPoint.getDirectDistance();
        
        
   
    }
  }
  
  ///////////////////////////////
  // GET METHODS               //
  ///////////////////////////////
  
  public String getGpx() {
    //<trkpt lat="47.676143" lon="15.723030"><ele>1070.737549</ele><time>2017-09-09T08:03:57Z</time></trkpt>
    String pGpx = "      <trkpt lat=\"" + 
      Utilities.formatDoubleEnglish(m_dLatitude) + "\" lon=\"" + 
      Utilities.formatDoubleEnglish(m_dLongitude) + "\"><ele>" + 
      Utilities.formatDoubleEnglish(m_dElevation) + "</ele><time>" + 
      Utilities.getDateAsString(m_pTimestamp) + "</time></trkpt>";
      /*
      System.out.println(pGpx);
      System.out.println(" <trkpt lat=\"" + m_dLatitude + "\" lon=\"" + 
      m_dLongitude + "\"><ele>" + 
      m_dElevation + "</ele><time>" + 
      m_pTimestamp + "</time></trkpt>");  */
    return pGpx;
  }
  
  public String toString() {
    return "Lat: " + m_dLatitude + ", Lon: " + m_dLongitude + ", Ele: " + m_dElevation + ", Time: " + getTimestampAsString();
  }
  
  public String getTimestampAsString() {
    return Utilities.getDateAsString(m_pTimestamp);
  }
  
  public String getTimeForHtml() {
    return Utilities.getTimeForHtml(m_pTimestamp);
  } 
  
  public long getTimeInMilliseconds() {
    return m_lTimeInMilliseconds;
  }
  
  public double getLatitude() {
    return m_dLatitude;
  }
  
  public double getLongitude() {
    return m_dLongitude;
  }
  
  public double getKilometers() {
    return m_dKilometers;
  }
  
  public double getDirectDistance() {
    return m_dDirectDistance;
  }
  
  public double getElevation() {
    return m_dElevation;
  }
  
  public String getHeightMapData(Target pPreviousTarget, Target pNextTarget) {
    String pCsvLine = "";
    
    double dKilometerToNext = 0;
    double dKilometerFromPrevious = 0;
    int iMinutesToNext = 0;
    int iMinutesFromPrevious = 0;
    double dElevationDiffToNext = 0;
    double dElevationDiffFromPrevious = 0;
    
    if (pPreviousTarget != null) {
      dKilometerFromPrevious = m_dKilometers - pPreviousTarget.getKilometers();
      iMinutesFromPrevious = Utilities.getAbsoluteMinuteDifference(getTimeForHtml(), pPreviousTarget.getTimeForHtml());
      dElevationDiffFromPrevious = m_dElevation - pPreviousTarget.getElevation();
    }
    
    if (pNextTarget != null) {
      dKilometerToNext = pNextTarget.getKilometers() - m_dKilometers;
      iMinutesToNext = Utilities.getAbsoluteMinuteDifference(getTimeForHtml(), pNextTarget.getTimeForHtml());
      dElevationDiffToNext = m_dElevation - pNextTarget.getElevation();
    }
    
    
    //height
    pCsvLine += m_dElevation + ";";
    //time
    pCsvLine += GetTimeForCsv() + ";";
    //gps kilometers
    pCsvLine += m_dKilometers + ";";
    //Kilometer bis zum nächsten Ziel
    pCsvLine += dKilometerToNext + ";";
    //Kilometer weg vom vorherigen Ziel
    pCsvLine += dKilometerFromPrevious + ";";
    //Minuten bis zum nächsten Ziel
    pCsvLine += iMinutesToNext + ";";
    //Minuten weg vom vorherigen Ziel
    pCsvLine += iMinutesFromPrevious + ";";
    //Höhendifferenz zu nächsten Ziel
    pCsvLine += dElevationDiffToNext + ";";
    //Höhendifferenz zu vorherigem Ziel
    pCsvLine += dElevationDiffFromPrevious + ";";
    
    
    return pCsvLine;
  }
  
  
  ////////////////////////////////////
  // PRIVATE GET METHODS            //
  ////////////////////////////////////
  private String GetTimeForCsv() {
    return TIME_FORMAT.format(m_pTimestamp);
  }
  
 
}